import pysam
import sys
from covernantlib.coveragecalculator import CoverageCalculator


def calc_ratio(args):
    coverage_ratio_calculator = CoverageRatioCalculator(args)
    coverage_ratio_calculator.set_names()
    coverage_ratio_calculator.calc_coverages()
    coverage_ratio_calculator.print_no_aligned_reads()
    coverage_ratio_calculator.calc_combined_factor()
    coverage_ratio_calculator.write_numerator_and_denominator_wiggle_files()
    coverage_ratio_calculator.compare()
    coverage_ratio_calculator.write_ratio_wiggle_files()


class CoverageRatioCalculator(object):

    def __init__(self, args):
        self._denominator_bam_file = args.denominator_bam_file
        self._numerator_bam_file = args.numerator_bam_file
        self._output_prefix = args.output_prefix
        if args.window_size is not None and args.window_size % 2 == 0:
            sys.stderr.write("Error. Window size must be an odd number!\n")
            sys.exit(2)
        self._window_size = args.window_size
        self._step_size = args.step_size
        self._factor = args.factor
        self._keep_zero_coverage = args.keep_zero_coverage
        self._denominator_name = args.denominator_name
        self._numerator_name = args.numerator_name
        self._ratio_name = args.ratio_name
        self._paired_end = args.paired_end

    def calc_coverages(self):
        self._print_file_names()
        (self.no_of_mapped_reads_numerator,
        self.no_of_mapped_reads_numerator_forward,
         self.no_of_mapped_reads_numerator_reverse
        ) = self._count_no_of_mapped_reads(
            self._numerator_bam_file)
        (self.no_of_mapped_reads_denominator,
         self.no_of_mapped_reads_denominator_forward,
         self.no_of_mapped_reads_denominator_reverse
        ) = self._count_no_of_mapped_reads(
            self._denominator_bam_file)
        (self.coverage_denominator,
         self.coverage_denominator_forward,
         self.coverage_denominator_reverse) = self._prepare_coverage(
             self._denominator_bam_file)
        (self.coverage_numerator,
         self.coverage_numerator_forward,
         self.coverage_numerator_reverse) = self._prepare_coverage(
             self._numerator_bam_file)

    def print_no_aligned_reads(self):
        print("Number of mapped reads in denominator sample - "
              "total: %s" % self.no_of_mapped_reads_denominator)
        print("Number of mapped reads in denominator sample - "
              "forward strand: %s" %
              self.no_of_mapped_reads_denominator_forward)
        print("Number of mapped reads in denominator sample - "
              "reverse strand: %s" %
              self.no_of_mapped_reads_denominator_reverse)
        print("Number of mapped reads in numerator sample total: %s" %
              self.no_of_mapped_reads_numerator)
        print("Number of mapped reads in numerator sample - "
              "forward strand: %s" % self.no_of_mapped_reads_numerator_forward)
        print("Number of mapped reads in numerator sample - "
              "reverse strand: %s" % self.no_of_mapped_reads_numerator_reverse)

    def calc_combined_factor(self):
        self._numerator_rpm_factor = 1000000.0/float(
            self.no_of_mapped_reads_numerator)
        self._denominator_rpm_factor = 1000000.0/float(
            self.no_of_mapped_reads_denominator)
        self._ratio_factor = float(self.no_of_mapped_reads_numerator)/float(
            self.no_of_mapped_reads_denominator)
        if self._factor is not None:
            self._ratio_factor *= self._factor
            print("Ratio factor: %s (%s/%s * %s) " % (
                self._ratio_factor, self.no_of_mapped_reads_numerator,
                self.no_of_mapped_reads_denominator, self._factor))
        else:
            print("Ratio factor: %s (%s/%s) " % (
                self._ratio_factor, self.no_of_mapped_reads_numerator,
                self.no_of_mapped_reads_denominator,))
        print("RPM factor numerator: %s (1M/%s)" % (
            self._numerator_rpm_factor, self.no_of_mapped_reads_numerator))
        print("RPM factor denominator: %s (1M/%s)" % (
            self._denominator_rpm_factor, self.no_of_mapped_reads_denominator))

    def set_names(self):
        if self._denominator_name is None:
            self._denominator_name = self._file_name_to_name(
                self._denominator_bam_file)
        if self._numerator_name is None:
            self._numerator_name = self._file_name_to_name(
                self._numerator_bam_file)
        if self._ratio_name is None:
            self._ratio_name = "{}_vs_{}".format(
                self._file_name_to_name(self._denominator_bam_file),
                self._file_name_to_name(self._numerator_bam_file))
        print(self._denominator_name, self._numerator_name, self._ratio_name)

    def _file_name_to_name(self, file_name):
        name = file_name.split("/")[-1].replace(".bam", "")
        return name

    def write_numerator_and_denominator_wiggle_files(self):
        self._write_wiggle(
            self._calc_averaged_coverages(self.coverage_denominator),
            "denominator_{}".format(self._denominator_name),
            self._denominator_rpm_factor)
        self._write_wiggle(
            self._calc_averaged_coverages(self.coverage_denominator_forward),
            "denominator_{}_forward".format(self._denominator_name),
            self._denominator_rpm_factor)
        self._write_wiggle(
            self._calc_averaged_coverages(self.coverage_numerator_forward),
            "numerator_{}_forward".format(self._numerator_name),
            self._numerator_rpm_factor)
        self._write_wiggle(
            self._calc_averaged_coverages(self.coverage_numerator),
            "numerator_{}".format(self._numerator_name),
            self._numerator_rpm_factor)
        self._write_wiggle(
            self._calc_averaged_coverages(self.coverage_denominator_reverse),
            "denominator_{}_reverse".format(self._denominator_name),
            self._denominator_rpm_factor)
        self._write_wiggle(
            self._calc_averaged_coverages(self.coverage_numerator_reverse),
            "numerator_reverse".format(self._numerator_name),
            self._numerator_rpm_factor)

    def _calc_averaged_coverages(self, coverages):
        averaged_coverages = {}
        # TODO - CHECK
        if self._window_size is None:
            self._window_size = 1
        for element, element_coverages in coverages.items():
            averaged_coverages[element] = self._sliding_windows_average(
                element_coverages)
        return averaged_coverages

    def write_ratio_wiggle_files(self):
        self._write_wiggle(
            self.elements_and_coverage_ratios,
            "ratio_{}".format(self._ratio_name), self._ratio_factor)
        self._write_wiggle(
            self.elements_and_coverage_ratios_forward,
            "ratio_{}_forward".format(self._ratio_name),
            self._ratio_factor)
        self._write_wiggle(
            self.elements_and_coverage_ratios_reverse,
            "ratio_{}_reverse".format(self._ratio_name),
            self._ratio_factor)

    def _write_wiggle(self, elements_and_coverages, name, factor):
        output_fh = open("%s-%s.wig" % (self._output_prefix, name), "w")
        output_fh.write("track type=wiggle_0 name=\"%s_%s\"\n" % (
            self._output_prefix.split("/")[-1], name))
        for element in sorted(elements_and_coverages.keys()):
            output_fh.write("variableStep chrom=%s span=1\n" % (element))
            # Remove position with as coverage of 0. pos is increased
            # by 1 as a translation from a 0-based sysem (Python list)
            # to a 1 based system (wiggle) takes place.
            output_fh.write(
                "\n".join(
                    ["%s %s" % (pos + 1, float(coverage) * factor)
                     for pos, coverage in
                     self._pos_and_coverages(elements_and_coverages[element])])
                + "\n")
        output_fh.close()

    def _pos_and_coverages(self, coverages):
        if not self._keep_zero_coverage:
            # As the values between the nucleotides selected by the
            # step size are zero only the selected nucleotides will be
            # returned due to this filter.
            for pos, coverage in filter(
                lambda pos_and_cov: pos_and_cov[1] != 0.0,
                    enumerate(coverages)):
                yield (pos, coverage)
        else:
            # Here the step size has to be used explicitly to return
            # the selected nucleotide
            for pos in range(int(self._window_size/2),
                             len(coverages) - int(self._window_size/2),
                             self._step_size):
                yield (pos, coverages[pos])

    def compare(self):
        self.elements_and_coverage_ratios = {}
        for element, coverages in self.coverage_denominator.items():
            self.elements_and_coverage_ratios[
                element] = self._compare_coverages(
                    element, self.coverage_denominator,
                    self.coverage_numerator)
        self.elements_and_coverage_ratios_forward = {}
        for element, coverages in self.coverage_denominator_forward.items():
            self.elements_and_coverage_ratios_forward[
                element] = self._compare_coverages(
                    element, self.coverage_denominator_forward,
                    self.coverage_numerator_forward)
        self.elements_and_coverage_ratios_reverse = {}
        for element, coverages in self.coverage_denominator_reverse.items():
            self.elements_and_coverage_ratios_reverse[
                element] = self._compare_coverages(
                    element, self.coverage_denominator_reverse,
                    self.coverage_numerator_reverse)

    def _compare_coverages(self, element, coverage_denominator,
                           coverage_numerator):
        cur_cov_denominator = coverage_denominator[element]
        cur_cov_numerator = coverage_numerator[element]
        if len(cur_cov_denominator) != len(cur_cov_numerator):
            sys.stderr.write("Error! Different number of nucleotides.\n")
            sys.exit(2)
        if self._window_size is not None:
            cur_cov_denominator = self._sliding_windows_average(
                cur_cov_denominator)
            cur_cov_numerator = self._sliding_windows_average(
                cur_cov_numerator)
        # Calculate the ratio of numerator data to denominator data
        coverage_ratios = [
            self._ratio(
                float(numerator) / float(self.no_of_mapped_reads_numerator),
                float(con) / float(self.no_of_mapped_reads_denominator))
            for numerator, con in zip(cur_cov_numerator, cur_cov_denominator)]
        return(coverage_ratios)

    def _prepare_coverage(self, bam_file):
        ref_seq_and_coverages_sum = {}
        ref_seq_and_coverages_forward = {}
        ref_seq_and_coverages_reverse = {}
        coverage_calculator = CoverageCalculator(
            read_count_splitting=False, uniqueley_aligned_only=False,
            first_base_only=False, paired_end=self._paired_end)
        for ref_seq, coverages in coverage_calculator.ref_seq_and_coverages(
                bam_file):
            assert len(coverages["forward"]) == len(coverages["reverse"])
            # Sum up the coverage of the forward and reverse strand
            summed_coverage = [
                abs(cov_for) + abs(cor_rev) for cov_for, cor_rev
                in zip(coverages["forward"], coverages["reverse"])]
            ref_seq_and_coverages_sum[ref_seq] = summed_coverage
            ref_seq_and_coverages_forward[ref_seq] = [
                abs(cor_forw) for cor_forw in coverages["forward"]]
            ref_seq_and_coverages_reverse[ref_seq] = [
                abs(cor_rev) for cor_rev in coverages["reverse"]]
        print("Number of used alignments: {}".format(
            coverage_calculator.used_alignmets))
        # print("Number of discarded alignments: {}".format(
        #     coverage_calculator.discared_alignments))
        return (ref_seq_and_coverages_sum, ref_seq_and_coverages_forward,
                ref_seq_and_coverages_reverse)

    def _print_file_names(self):
        print("Performing ChIP-Seq analysis")
        print("- Input file with numerator data: %s" %
              self._numerator_bam_file)
        print("- Input file with denominator data: %s" %
              self._denominator_bam_file)
        print("- Output file prefix: %s" % self._output_prefix)

    def _sliding_windows_average(self, coverages):
        averaged_coverages = [0] * len(coverages)
        for pos in range(int(self._window_size/2),
                         len(coverages) - int(self._window_size/2),
                         self._step_size):
            averaged_coverages[pos] = (
                sum(coverages[pos-int(self._window_size/2):
                              pos+int(self._window_size/2)+1])
                / self._window_size)
        return(averaged_coverages)

    def _count_no_of_mapped_reads(self, bam_file):
        reads_forward = set()
        reads_reverse = set()
        with pysam.Samfile(bam_file, "rb") as bam_fh:
            for read in bam_fh.fetch():
                if read.is_reverse is False:
                    reads_forward.add(read.qname)
                else:
                    reads_reverse.add(read.qname)
        return(len(reads_forward) + len(reads_reverse), len(reads_forward),
               len(reads_reverse))

    def _ratio(self, mult, div):
        try:
            return(mult/div)
        except:
            return(0.0)



import pysam


class CoverageCalculator(object):

    def __init__(self, read_count_splitting=True, uniqueley_aligned_only=False,
                 first_base_only=False, paired_end=False):
        self._read_count_splitting = read_count_splitting
        self._uniqueley_aligned_only = uniqueley_aligned_only
        self._first_base_only = first_base_only
        self._coverage_add_function = self._select_coverage_add_function()
        self._coverages = {}
        self._paired_end = paired_end
        self.used_alignmets = 0

    def ref_seq_and_coverages(self, bam_path):
        bam = self._open_bam_file(bam_path)
        for ref_seq, length in zip(bam.references, bam.lengths):
            self._init_coverage_list(length)
            self._calc_coverage(ref_seq, bam)
            yield(ref_seq, self._coverages)

    def _init_coverage_list(self, length):
        for strand in ["forward", "reverse"]:
            self._coverages[strand] = [0.0] * length

    def _calc_coverage(self, ref_seq, bam):
        if not self._paired_end:
            self._calc_coverage_single_end(ref_seq, bam)
        else:
            self._calc_coverage_paired_end(ref_seq, bam)
            
    def _calc_coverage_single_end(self, ref_seq, bam):
        for alignment in bam.fetch(ref_seq):
            self._add_coverage_of_single_end_reads(alignment)
            
    def _add_coverage_of_single_end_reads(self, alignment):
        number_of_hits = self._number_of_hits(alignment)
        if self._uniqueley_aligned_only is True and number_of_hits != 1:
            return
        # Note: No translation from bam coordinates to python
        # list coorindates is needed.
        start = alignment.pos
        end = alignment.aend
        increment = self._calc_increment_values(number_of_hits)
        if end is None or start is None:
            return
        self.used_alignmets += 1
        self._coverage_add_function(alignment, increment, start, end)

    def _calc_coverage_paired_end(self, ref_seq, bam):
        """Can handle single and paired end reads. In case of paired end
        reads the reads are treates as one fragment."""
        read_pairs_by_qname = {}
        for alignment in bam.fetch(ref_seq):
            if not alignment.is_proper_pair:
                self._add_coverage_of_single_end_reads(alignment)
            else:
                # Collect all read pairs - this might be memory intensive!
                index = 0
                if alignment.is_read2:
                    index = 1
                read_pairs_by_qname.setdefault(
                    alignment.qname, [None, None])
                read_pairs_by_qname[alignment.qname][index] = alignment
        for pair in read_pairs_by_qname.values():
            # TO FIX - raises Segementation fault:
            # if None in pair:
            #    continue
            read_1, read_2 = pair
            try:
                start = min(read_1.pos, read_1.aend, read_2.pos, read_2.aend)
            except AttributeError:
                continue
            end = max(read_1.pos, read_1.aend, read_2.pos, read_2.aend)
            number_of_hits = self._number_of_hits(read_1)
            increment = self._calc_increment_values(number_of_hits)
            self._coverage_add_function(read_1, increment, start, end)
            self.used_alignmets += 1

    def _number_of_hits(self, alingment):
            try:
                return dict(alingment.tags)["NH"]
            except KeyError:
                return 1

    def _calc_increment_values(self, number_of_hits):
        # Normalize coverage increment by number of read alignments
        # per read
        if self._read_count_splitting is True:
            return 1.0 / float(number_of_hits)
        else:
            return 1.0
            
    def _select_coverage_add_function(self):
        if self._first_base_only is False:
            return self._add_whole_alignment_coverage
        else:
            return self._add_first_base_coverage

    def _open_bam_file(self, bam_file):
        return pysam.AlignmentFile(bam_file)

    def _close_bam_fh(self, bam_fh):
        bam_fh.close()

    def _add_whole_alignment_coverage(self, alignment, increment, start, end):
        if alignment.is_reverse is False:
            self._coverages["forward"][start:end] = [
                coverage + increment for coverage in
                self._coverages["forward"][start:end]]
        else:
            self._coverages["reverse"][start:end] = [
                coverage - increment for coverage in
                self._coverages["reverse"][start:end]]

    def _add_first_base_coverage(self, alignment, increment, start, end):
        if alignment.is_reverse is False:
            self._coverages["forward"][start] = self._coverages[
                "forward"][start] + increment
        else:
            self._coverages["reverse"][end-1] = self._coverages[
                "reverse"][end-1] - increment




import csv
from datetime import datetime
import sys
import math
import numpy as np
import pandas as pd
from covernantlib.wiggle import WiggleParser


def extract(args):
    coverage_extractor = CoverageExtractor(
        args.coordinate_file, args.coverage_file, args.output_prefix,
        args.flip_reverse_strand, args.matrix_alignment, args.window_size,
        args.step_size)
    log("Reading wiggle file")
    coverage_extractor.read_wiggle_file()
    log("Reading coordinate files")
    coverage_extractor.read_coordinate_file()
    log("Generate coverage matrix")
    coverage_extractor.extract_coverages()
    coverage_extractor.generate_coverage_matrix()
    log("Writing matrix")
    coverage_extractor.write_matrix_to_file()
    log("Calculating and writing averages")
    coverage_extractor.calc_averages()

    
def log(msg):
    sys.stderr.write("{} - {}\n".format(
        datetime.strftime(datetime.now(), '%Y-%m-%d %H:%M:%S'), msg))


class CoverageExtractor(object):
    """ """
    def __init__(self, coordinate_file, coverage_file, output_prefix,
                 flip_reverse_strand, matrix_alignment, window_size,
                 step_size):
        self._coordinate_file = coordinate_file
        self._coverage_file = coverage_file
        self._output_prefix = output_prefix
        self._flip_reverse_strand = flip_reverse_strand
        self._matrix_alignment = matrix_alignment
        self._replicons_and_coverages = {}
        self._coordinates = []
        self._coverage_lists = []
        self._window_size = window_size
        self._step_size = step_size

    def read_wiggle_file(self):
        wiggle_parser = WiggleParser()
        for entry in wiggle_parser.entries(open(self._coverage_file)):
            # Return dictionaries of the coordinate and absolute
            # coverage values
            self._replicons_and_coverages[entry.replicon] = dict(
                [(pos, abs(value)) for pos, value in entry.pos_value_pairs])

    def read_coordinate_file(self):
        for row in csv.reader(open(self._coordinate_file), delimiter="\t"):
            if row[0].startswith("#"):
                continue
            start, end = [int(pos) for pos in row[1:3]]
            assert start <= end
            strand = row[3]
            assert strand in ["+", "-"]
            self._coordinates.append(
                {"replicon": row[0], "start": start,
                 "end": end, "strand": strand})

    def extract_coverages(self):
        for coordinate in self._coordinates:
            self._check_coordinate_validity(coordinate)
            coverages = [
                self._replicons_and_coverages[coordinate["replicon"]].get(
                    pos, 0.0)
                for pos in range(coordinate["start"], coordinate["end"])
            ]
            if self._flip_reverse_strand and coordinate["strand"] == "-":
                coverages = coverages[::-1]
            self._coverage_lists.append(coverages)

    def _check_coordinate_validity(self, coordinate):
        pass
            
    def generate_coverage_matrix(self):
        max_range = max(
            [len(coverages) for coverages in self._coverage_lists])
        self._coverage_df = self._init_coverage_dataframe(max_range)
        for coordinate, coverages in zip(
                self._coordinates, self._coverage_lists):
            aligned_coverages = self._align_coverages(coverages, max_range)
            if self._window_size == 1 and self._step_size == 1:
                row = [coordinate[key] for key in
                       ["replicon", "start", "end",
                        "strand"]] + aligned_coverages
            else:
                row = [coordinate[key] for key in
                       ["replicon", "start", "end", "strand"]]
                for pos in range(int(self._window_size/2) + 1,
                                 max_range - int(self._window_size/2) + 1,
                                 self._step_size):
                    start = pos-int(self._window_size/2)
                    end = pos+int(self._window_size/2)
                    coverages = aligned_coverages[start:end]
                    row.append(np.mean(coverages))
            row_df = self._init_coverage_dataframe(max_range, [row])
            self._coverage_df = self._coverage_df.append(
                row_df, ignore_index=True)

    def _align_coverages(self, coverages, max_range):
        filling = np.nan
        aligned_coverages = list(np.full(max_range, filling))
        if self._matrix_alignment == "left":
            aligned_coverages[:len(coverages)] = coverages
        elif self._matrix_alignment == "center":
            margin = round((max_range - len(coverages)) / 2)
            aligned_coverages[margin:margin+len(coverages)] = coverages
        elif self._matrix_alignment == "right":
            left_margin = max_range - len(coverages)
            aligned_coverages[
                left_margin:left_margin+len(coverages)] = coverages
        return aligned_coverages
            
    def _init_coverage_dataframe(self, max_range, data=None):
        """
        - max_range
        - data - list of list containig data
        """
        return pd.DataFrame(
            data, columns=["Replicon", "Start", "End", "Strand"] +
            list(range(int(self._window_size/2) + 1,
                       max_range - int(self._window_size/2) + 1,
                       self._step_size)))

    def write_matrix_to_file(self):
        self._coverage_df.to_csv(
            "{}_matrix.csv".format(self._output_prefix),
            sep="\t", index=False, na_rep="NaN")

    def calc_averages(self):
        average_df = pd.DataFrame(columns=["Value"] +
                                  list(self._coverage_df.columns)[4:])
        average_df["Value"] = [
            "Median (all values)",
            "Mean (all values)",
            "Standard deviation (all values)",
            "Median (zeros removed)",
            "Mean (zeros removed)",
            "Standard deviation (zeros removed)",
            "Median (NAs removed)",
            "Mean (NAs removed)",
            "Standard deviation (NAs removed)",
            "Median (zeroes and NAs removed)",
            "Mean (zeroes and NAs removed)",
            "Standard deviation (zeroes and NAs removed)"
            ]
        for column_name in list(self._coverage_df.columns)[4:]:
            column = self._coverage_df[column_name]
            column_without_zeros = np.array(list(
                filter(lambda cov: cov != 0.0, column)))
            column_without_na = np.array(list(
                filter(lambda cov: not math.isnan(cov), column)))
            column_without_zeros_and_na = np.array(list(filter(
                lambda cov: not (math.isnan(cov) or cov == 0.0), column)))
            average_df[column_name] = [
                column.median(),
                column.mean(),
                column.std(),
                self._save_column_operation(np.median, column_without_zeros),
                self._save_column_operation(np.mean, column_without_zeros),
                self._save_column_operation(np.std, column_without_zeros),
                self._save_column_operation(np.median, column_without_na),
                self._save_column_operation(np.mean, column_without_na),
                self._save_column_operation(np.std, column_without_na),
                self._save_column_operation(
                    np.median, column_without_zeros_and_na),
                self._save_column_operation(
                    np.mean, column_without_zeros_and_na),
                self._save_column_operation(
                    np.std, column_without_zeros_and_na)
            ]
        average_df.to_csv(
            "{}_averages.csv".format(self._output_prefix),
            sep="\t", index=False)

    def _save_column_operation(self, function, column):
        if len(column) > 0:
            return function(column)
        else:
            return "-"



class WiggleParser(object):
    """

    Warning - this does not implement the full specification!

    """

    def entries(self, input_fh):
        track_name = None
        replicon = None
        span = None
        pos_value_pairs = []
        for line in input_fh:
            row = line[:-1].split()
            if len(row) == 0:
                continue
            if row[0].startswith("track"):
                track_name = self._track_name(row)
            elif row[0].startswith("variableStep"):
                if replicon:
                    prev_replicon = replicon
                    prev_span = span
                    prev_pos_value_pairs = pos_value_pairs
                    replicon = self._replicon(row)
                    span = None
                    pos_value_pairs = []
                    yield WiggleEntry(
                            track_name, prev_replicon, prev_span, 
                            prev_pos_value_pairs)
                else:
                    replicon = self._replicon(row)
            else:
                pos_value_pairs.append([int(row[0]), float(row[1])])
        yield WiggleEntry(track_name, replicon, span, pos_value_pairs)

    def _replicon(self, row):
        return self._attrs_and_values(row)["chrom"]

    def _track_name(self, row):
        return self._attrs_and_values(row)["name"]

    def _attrs_and_values(self, row):
        attrs_and_values = {}
        for attr_and_value in row:
            if not "=" in attr_and_value:
                continue
            attr, value = attr_and_value.split("=")
            value = value.replace("\"", "")
            attrs_and_values[attr] = value
        return attrs_and_values


class WiggleEntry(object):

    def __init__(self, track_name, replicon, span, pos_value_pairs):
        self.track_name = track_name
        self.replicon = replicon
        self.span = span
        self.pos_value_pairs = pos_value_pairs


class WiggleWriter(object):

    def __init__(self, track_str, fh):
        self._fh = fh
        self._fh.write(("track type=wiggle_0 name=\"%s\"\n" % (track_str)))

    def write_replicons_coverages(
            self, replicon, pos_value_pairs, factor=1.0):
            self._fh.write("variableStep chrom=%s span=1\n" % (replicon))
            # Filter values of 0 and multiply the remaining ones by
            # the given factor.
            self._fh.write(
                "\n".join(["%s %s" % (pos, coverage * factor)
                           for pos, coverage in
                           filter(lambda pos_and_cov: pos_and_cov[1] != 0.0,
                                  pos_value_pairs)]) + "\n")

    def close_file(self):
        self._fh.close()



import sys
from datetime import datetime
import math
import pandas as pd
import numpy as np
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages


def plot_matrix(args):
    pp = PdfPages(args.output_file)
    matrix = pd.read_table(args.matrix_file)
    if args.share_y_max:
        max_y = np.nanmax([
            np.nanmax(row[4:]) for row in map(list, matrix.values)])
    x_values = [int(x) for x in list(matrix.columns)[4:]]
    if args.share_x_range:
        min_x = np.nanmin(x_values)
        max_x = np.nanmax(x_values)
    for row in map(list, matrix.values):
        range_name = "{} {} {} {}".format(row[0], row[1], row[2], row[3])
        log("Plotting {}".format(range_name))
        y_values = row[4:]
        assert len(x_values) == len(y_values)
        filtered_x_values = []
        filtered_y_values = []
        for x, y in zip(x_values, y_values):
            if not math.isnan(y):
                filtered_x_values.append(x)
                filtered_y_values.append(y)
        if args.share_x_range:
            cur_min_x = min_x
            cur_max_x = max_x
        else:
            cur_min_x = np.nanmin(filtered_x_values)
            cur_max_x = np.nanmax(filtered_x_values)
        cur_max_y = max_y if args.share_y_max else np.nanmax(filtered_y_values)
        fig = plt.figure()
        plt.plot(x_values, y_values, color="black", )
        plt.title(range_name)
        plt.axis([cur_min_x, cur_max_x, 0, cur_max_y])
        plt.xlabel("Position")
        plt.ylabel("Coverage")
        pp.savefig()
        plt.close(fig)
    pp.close()

    
def log(msg):
    sys.stderr.write("{} - {}\n".format(
        datetime.strftime(datetime.now(), '%Y-%m-%d %H:%M:%S'), msg))



def rescale_wiggle(args):
    with open(args.output_prefix + ".wig", "w") as output_fh:
        with open(args.input_file) as input_fh:
            for line in input_fh:
                if not (line.startswith("variableStep")
                        or line.startswith("track")
                        or len(line.strip()) == 0):
                    pos, value = line.split()
                    line = "{} {}\n".format(
                        pos, float(value) * args.factor)
                output_fh.write(line)



import pybedtools
from collections import defaultdict
import numpy as np


def bed_to_wiggle(input_file, output_prefix, window_size, step_size):
    for strand in ["+", "-"]:
        _generate_coverage_for_strand(
            strand, input_file, output_prefix, window_size, step_size)


def _generate_coverage_for_strand(strand, input_file, output_prefix,
                                  window_size, step_size):
    replicons_and_coverages = init_coverage_dict(input_file)
    entries = pybedtools.BedTool(input_file)
    for replicon, coverages in replicons_and_coverages.items():
        for window_center in range(
                int(window_size/2),
                len(coverages)-int(window_size/2), step_size):
            interval = pybedtools.create_interval_from_list(
                [replicon, window_center-int(window_size/2),
                 window_center+int(window_size/2), "tmp_window", ".", strand])
            factor = 1
            if strand == "-":
                factor = -1
            replicons_and_coverages[replicon][
                window_center] = factor * entries.count_hits(
                    interval, same_strand=True)

    output_fh = open("{}_{}_strand.wig".format(
        output_prefix, strand), "w")
    output_fh.write("track type=wiggle_0 name=\"coverage {} strand\"\n".format(
        strand))
    
    for replicon in sorted(replicons_and_coverages.keys()):
        output_fh.write("variableStep chrom=%s span=1\n" % (replicon))
        # Remove position with as coverage of 0. pos is increased
        # by 1 as a translation from a 0-based sysem (Python list)
        # to a 1 based system (wiggle) takes place.
        output_fh.write(
            "\n".join(
                ["%s %s" % (pos + 1, coverage)
                 for pos, coverage in enumerate(
                         replicons_and_coverages[replicon])]) + "\n")
    output_fh.close()


def init_coverage_dict(input_file):
    replicons_and_max_end_pos = defaultdict(int)
    for entry in pybedtools.BedTool(input_file):
        if entry.end > replicons_and_max_end_pos[entry.chrom]:
            replicons_and_max_end_pos[entry.chrom] = entry.end
    replicons_and_coverages = {}
    for replicon, end_pos in replicons_and_max_end_pos.items():
        replicons_and_coverages[replicon] = np.zeros(end_pos)
    return replicons_and_coverages
